#include "version.hpp"
#include "config.hpp"
#include "service_runner.hpp"
#include "services.hpp"
#include "servers.hpp"
#include "utils/directories.hpp"
#include "templates.hpp"
#include "utils/utils.hpp"
#include <boost/filesystem.hpp>

#include <iostream>
#include <string>
#include <vector>
#include <iomanip>
namespace dropshell {

void print_help() {
    std::cout << std::endl;
    maketitle("DropShell version " + VERSION);
    std::cout << std::endl;
    std::cout << "A tool for managing server configurations" << std::endl;
    std::cout << std::endl;
    std::cout << "dropshell ..." << std::endl;
    std::cout << "  help           Show this help message" << std::endl;
    std::cout << "  init DIR       Add DIR as a local configuration directory for dropshell (can add several)" << std::endl;
    std::cout << "  server NAME    Show details for specific server" << std::endl;
    std::cout << "  templates      List all available templates" << std::endl;
    std::cout << std::endl;
    std::cout << std::endl;
    std::cout << "Service commands: (if no service is specified, all services for the server are affected)" << std::endl;
    std::cout << "  install SERVER [SERVICE]     Install/reinstall/update service(s). Non-destructive." << std::endl;
    std::cout << "  COMMAND SERVER [SERVICE]     Run a command on service(s)." << std::endl;
    std::cout << std::endl;
    std::cout << "Standard commands: install, uninstall, backup, restore, start, stop" << std::endl;
    std::cout << std::endl;
    std::cout << "  ssh     SERVER [SERVICE]     Launch an interactive shell on a server or service" << std::endl;
    std::cout << std::endl;
    std::cout << "Creation commands: (apply to the first local config directory)"<<std::endl;
    std::cout << "  create-template TEMPLATE" << std::endl;
    std::cout << "  create-server   SERVER" << std::endl;
    std::cout << "  create-service  SERVER SERVICE" << std::endl;
}

} // namespace dropshell


bool parseargs(std::string arg2, std::string arg3, std::string & server_name, std::vector<dropshell::ServiceInfo>& servicelist)
{
    if (arg2.empty()) return false;
    server_name = arg2;

    if (arg3.empty()) {
        servicelist = dropshell::get_server_services_info(server_name);
    } else {
        servicelist.push_back(dropshell::get_service_info(server_name, arg3));
    }

    return true;
}

std::string safearg(int argc, char *argv[], int index)
{
    if (index >= argc) return "";
    return argv[index];
}

int main(int argc, char* argv[]) {
    try {
        dropshell::config *cfg = dropshell::get_global_config();
        // silently attempt to load the config file.
        cfg->load_config();

        if (argc == 1) {
            dropshell::list_servers();
            std::cout << std::endl;
            std::cout << "dropshell help   - get help on using dropshell" << std::endl;
            return 0;
        }
        std::string cmd = argv[1];

        // don't load old config if we're initializing
        if (cmd == "init") {
            std::string lcd;

            if (argc < 3) {
                std::cerr << "Error: init command requires a directory argument" << std::endl;
                return 1;
            }
            try {
                if (!cfg->add_local_config_directory(argv[2])) 
                    return 1; // error already reported
                cfg->save_config();
                std::cout << "Config directory added: " << cfg->get_local_config_directories().back() << std::endl;
                if (cfg->get_local_config_directories().size() ==1)
                    std::cout << "DropShell is now initialised and you can add a server with 'dropshell create-server <server-name>'" << std::endl;
                else
                {
                    std::cout << "DropShell will now use all of the following directories for configuration:" << std::endl;
                    for (const auto& dir : cfg->get_local_config_directories()) {
                        std::cout << "  " << dir << std::endl;
                    }
                    std::cout << "You can edit the config file manually at: " << dropshell::get_local_dropshell_config_path() << std::endl;
                }
                return 0;
            } catch (const std::exception& e) {
                std::cerr << "Error in init: " << e.what() << std::endl;
                return 1;
            }
        }

        if (cmd == "help" || cmd == "-h" || cmd == "--help" || cmd== "h" || cmd=="halp") {
            dropshell::print_help();
            return 0;
        }

        // auto completion stuff.
        std::set<std::string> commands;
        std::vector<dropshell::ServerInfo> servers = dropshell::get_configured_servers();
        for (const auto& server : servers) 
        {
            std::vector<dropshell::ServiceInfo> services = dropshell::get_server_services_info(server.name);
            for (const auto& service : services) 
                commands.merge(dropshell::get_used_commands(server.name, service.service_name));
        }

        if (cmd == "autocomplete_list_commands") { 
            commands.merge(std::set<std::string>{
                "help","init"
            });
            if (cfg->is_config_set()) 
                commands.merge(std::set<std::string>{
                    "server","templates","create-service","create-template","create-server","ssh"
                });
            
            for (const auto& command : commands) {
                std::cout << command << std::endl;
            }
            return 0;
        }

        if (cmd == "autocomplete_list_servers") {
            if (cfg->is_config_set()) 
            {
                auto servers = dropshell::get_configured_servers();
                for (const auto& server : servers) 
                    std::cout << server.name << std::endl;
            }
            return 0;
        }

        if (cmd == "autocomplete_list_services") {
            if (argc < 3) {
                std::cerr << "Error: autocomplete_list_services requires a server name" << std::endl;
                return 1;
            }
            if (cfg->is_config_set())  {
                auto services = dropshell::get_server_services_info(argv[2]);
                for (const auto& service : services) 
                    std::cout << service.service_name << std::endl;
            }
            return 0;
        }

        // ------------------------------------------------------------
        // from here we require the config file to be loaded.
        if (!cfg->is_config_set()) {   
            std::cerr << "Please run 'dropshell init <path>' to initialise the user directory and create a configuration file." << std::endl;
            return 1;
        }

        std::string lcd;    
        const std::vector<std::string> & local_config_directories = cfg->get_local_config_directories();
        std::cout << "Config directories: ";
        for (auto & dir : local_config_directories) 
            std::cout << "["<< dir << "] " << std::endl;
        std::cout << std::endl;;
        // No arguments provided
        if (argc < 2) {
            dropshell::print_help();
            return 0;
        }

        if (cmd == "server") {
            if (argc > 2) {
                // Show details for specific server
                dropshell::show_server_details(argv[2]);
            } else {
                // List all servers
                dropshell::list_servers();
            }
            return 0;
        }

        if (cmd == "templates") {
            dropshell::list_templates();
            return 0;
        }

        if (cmd == "create-template") {
            if (argc < 3) {
                std::cerr << "Error: create-template requires a template name" << std::endl;
                return 1;
            }
            dropshell::create_template(argv[2]);
            return 0;
        }

        if (cmd == "create-server") {
            if (argc < 3) {
                std::cerr << "Error: create-server requires a server name" << std::endl;
                return 1;
            }
            dropshell::create_server(argv[2]);
            return 0;
        }

        if (cmd == "create-service") {
            if (argc < 4) {
                std::cerr << "Error: create-service requires a server name and service name" << std::endl;
                return 1;
            }
            dropshell::create_service(argv[2], argv[3]);
            return 0;
        }

        if (cmd == "ssh" && argc < 4) {
            if (argc < 3) 
            {
                std::cerr << "Error: ssh requires a server name and optionally service name" << std::endl;
                return 1;
            }
            dropshell::interactive_ssh(argv[2], "");
            return 0;
        }

        // handle running a command. 
        for (const auto& command : commands) {
            if (cmd == command) {
                std::string server_name;
                std::vector<dropshell::ServiceInfo> servicelist;
                if (!parseargs(safearg(argc, argv, 2), safearg(argc, argv, 3), server_name, servicelist)) {
                    std::cerr << "Error: " << command << " command requires server name and optionally service name" << std::endl;
                    return 1;
                }

                for (const auto& service_info : servicelist) {
                    dropshell::service_runner service;
                    if (!service.init(server_name, service_info.service_name)) {
                        std::cerr << "Error: Failed to initialize service" << std::endl;
                        return 1;
                    }
                    if (!service.run_command(command)) {
                        std::cerr << command +" failed." << std::endl;
                        return 1;
                    }
                }
                return 0;
            }
        }

        // Unknown command
        std::cerr << "Error: Unknown command '" << cmd << "'" << std::endl;
        dropshell::print_help();
        return 1;

    } catch (const std::exception& e) {
        std::cerr << "Error: " << e.what() << std::endl;
        return 1;
    }
} 